/*
  Copyright (c) 2011 Alex Diener
  
  This software is provided 'as-is', without any express or implied
  warranty. In no event will the authors be held liable for any damages
  arising from the use of this software.
  
  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:
  
  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.
  
  Alex Diener adiener@sacredsoftware.net
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#if defined(__APPLE__)
#include <GLUT/glut.h>
#include <OpenGL/OpenGL.h>
#else
#include <GL/glut.h>
#endif

#if defined(__APPLE__)
#include <mach/mach.h>
#include <mach/mach_time.h>
#elif defined(__linux)
#include <time.h>
#elif defined(WIN32)
#include <windows.h>
#endif

#include "CollisionSystem.h"
#include "FixedIntervalRunLoop.h"

#define NUM_STATIC_OBJECTS 9

static FixedIntervalRunLoop runLoop;
static unsigned int windowWidth = 800, windowHeight = 600;
static struct movingObject movingObject;
static struct staticObject staticObjects[NUM_STATIC_OBJECTS];

double getCurrentTime() {
#if defined(__APPLE__)
	static mach_timebase_info_data_t timebaseInfo;
	
	if (timebaseInfo.denom == 0) {
		mach_timebase_info(&timebaseInfo);
	}
	return mach_absolute_time() * (double) timebaseInfo.numer / timebaseInfo.denom * 0.000000001;
	
#elif defined(WIN32)
	static LARGE_INTEGER frequency;
	LARGE_INTEGER currentTime;
	
	if (frequency.QuadPart == 0) {
		QueryPerformanceFrequency(&frequency);
	}
	QueryPerformanceCounter(&currentTime);
	
	return (double) currentTime.QuadPart / frequency.QuadPart;
	
#elif defined(__linux)
	struct timespec currentTime;
	
	clock_gettime(CLOCK_MONOTONIC, &currentTime);
	return currentTime.tv_sec + currentTime.tv_nsec * 0.000000001;
#endif
}

#define ACCELERATION 0.0078125f
#define DECELERATION 0.015625f
#define VELOCITY_MAX 0.125f

static void run() {
	if (movingObject.moveX == 0) {
		if (movingObject.velocity.x < 0.0f) {
			movingObject.velocity.x += DECELERATION;
			if (movingObject.velocity.x > 0.0f) {
				movingObject.velocity.x = 0.0f;
			}
			
		} else if (movingObject.velocity.x > 0.0f) {
			movingObject.velocity.x -= DECELERATION;
			if (movingObject.velocity.x < 0.0f) {
				movingObject.velocity.x = 0.0f;
			}
		}
		
	} else {
		if (movingObject.velocity.x * movingObject.moveX < 0) {
			movingObject.velocity.x += DECELERATION * movingObject.moveX;
		} else {
			movingObject.velocity.x += ACCELERATION * movingObject.moveX;
		}
		if (movingObject.velocity.x < -VELOCITY_MAX) {
			movingObject.velocity.x = -VELOCITY_MAX;
		} else if (movingObject.velocity.x > VELOCITY_MAX) {
			movingObject.velocity.x = VELOCITY_MAX;
		}
	}
	
	if (movingObject.moveY == 0) {
		if (movingObject.velocity.y < 0.0f) {
			movingObject.velocity.y += DECELERATION;
			if (movingObject.velocity.y > 0.0f) {
				movingObject.velocity.y = 0.0f;
			}
			
		} else if (movingObject.velocity.y > 0.0f) {
			movingObject.velocity.y -= DECELERATION;
			if (movingObject.velocity.y < 0.0f) {
				movingObject.velocity.y = 0.0f;
			}
		}
		
	} else {
		if (movingObject.velocity.y * movingObject.moveY < 0) {
			movingObject.velocity.y += DECELERATION * movingObject.moveY;
		} else {
			movingObject.velocity.y += ACCELERATION * movingObject.moveY;
		}
		if (movingObject.velocity.y < -VELOCITY_MAX) {
			movingObject.velocity.y = -VELOCITY_MAX;
		} else if (movingObject.velocity.y > VELOCITY_MAX) {
			movingObject.velocity.y = VELOCITY_MAX;
		}
	}
	
	movingObject.lastPosition = movingObject.position;
	movingObject.position.x += movingObject.velocity.x;
	movingObject.position.y += movingObject.velocity.y;
	
	resolveCollisions(&movingObject, staticObjects, NUM_STATIC_OBJECTS);
}

static void initObjects() {
	// Player
	movingObject.moveX = 0;
	movingObject.moveY = 0;
	movingObject.position.x = -0.5f;
	movingObject.position.y = -0.5f;
	movingObject.velocity.x = 0.0f;
	movingObject.velocity.y = 0.0f;
	movingObject.width = 1.0f;
	movingObject.height = 1.0f;
	
	// Outer walls
	staticObjects[0].position.x = -16.0f;
	staticObjects[0].position.y = -12.0f;
	staticObjects[0].width = 0.5f;
	staticObjects[0].height = 24.0f;
	staticObjects[1].position.x = 15.5f;
	staticObjects[1].position.y = -12.0f;
	staticObjects[1].width = 0.5f;
	staticObjects[1].height = 24.0f;
	staticObjects[2].position.x = -16.0f;
	staticObjects[2].position.y = -12.0f;
	staticObjects[2].width = 32.0f;
	staticObjects[2].height = 0.5f;
	staticObjects[3].position.x = -16.0f;
	staticObjects[3].position.y = 11.5f;
	staticObjects[3].width = 32.0f;
	staticObjects[3].height = 0.5f;
	
	// Solitary block
	staticObjects[4].position.x = 4.0f;
	staticObjects[4].position.y = -3.0f;
	staticObjects[4].width = 4.0f;
	staticObjects[4].height = 3.0f;
	
	// Vertical pair with small gap
	staticObjects[5].position.x = -10.0f;
	staticObjects[5].position.y = -1.75f;
	staticObjects[5].width = 1.75f;
	staticObjects[5].height = 1.75f;
	staticObjects[6].position.x = -10.0f;
	staticObjects[6].position.y = 0.0f;
	staticObjects[6].width = 1.75f;
	staticObjects[6].height = 1.75f;
	
	// Horizontal pair with small gap
	staticObjects[7].position.x = -1.75f;
	staticObjects[7].position.y = -8.0f;
	staticObjects[7].width = 1.75f;
	staticObjects[7].height = 1.75f;
	staticObjects[8].position.x = 0.0f;
	staticObjects[8].position.y = -8.0f;
	staticObjects[8].width = 1.75f;
	staticObjects[8].height = 1.75f;
}

#define SCALE 12.0f
#define STATIC_OUTLINE_RADIUS 0.0625f

static void displayFunc() {
	float ratio;
	size_t objectIndex;
	
	FixedIntervalRunLoop_run(&runLoop);
	
	glClear(GL_COLOR_BUFFER_BIT);
	
	glLoadIdentity();
	ratio = (float) windowWidth / (float) windowHeight;
	glOrtho(-ratio * SCALE, ratio * SCALE, -SCALE, SCALE, -1.0f, 1.0f);
	
	glBegin(GL_QUADS);
	
	for (objectIndex = 0; objectIndex < NUM_STATIC_OBJECTS; objectIndex++) {
		glColor4f(1.0f, 0.675f, 0.25f, 1.0f);
		glVertex2f(staticObjects[objectIndex].position.x, staticObjects[objectIndex].position.y);
		glVertex2f(staticObjects[objectIndex].position.x + STATIC_OUTLINE_RADIUS, staticObjects[objectIndex].position.y);
		glVertex2f(staticObjects[objectIndex].position.x + STATIC_OUTLINE_RADIUS, staticObjects[objectIndex].position.y + staticObjects[objectIndex].height);
		glVertex2f(staticObjects[objectIndex].position.x, staticObjects[objectIndex].position.y + staticObjects[objectIndex].height);
		
		glVertex2f(staticObjects[objectIndex].position.x + staticObjects[objectIndex].width - STATIC_OUTLINE_RADIUS, staticObjects[objectIndex].position.y);
		glVertex2f(staticObjects[objectIndex].position.x + staticObjects[objectIndex].width, staticObjects[objectIndex].position.y);
		glVertex2f(staticObjects[objectIndex].position.x + staticObjects[objectIndex].width, staticObjects[objectIndex].position.y + staticObjects[objectIndex].height);
		glVertex2f(staticObjects[objectIndex].position.x + staticObjects[objectIndex].width - STATIC_OUTLINE_RADIUS, staticObjects[objectIndex].position.y + staticObjects[objectIndex].height);
		
		glVertex2f(staticObjects[objectIndex].position.x, staticObjects[objectIndex].position.y);
		glVertex2f(staticObjects[objectIndex].position.x + staticObjects[objectIndex].width, staticObjects[objectIndex].position.y);
		glVertex2f(staticObjects[objectIndex].position.x + staticObjects[objectIndex].width, staticObjects[objectIndex].position.y + STATIC_OUTLINE_RADIUS);
		glVertex2f(staticObjects[objectIndex].position.x, staticObjects[objectIndex].position.y + STATIC_OUTLINE_RADIUS);
		
		glVertex2f(staticObjects[objectIndex].position.x, staticObjects[objectIndex].position.y + staticObjects[objectIndex].height - STATIC_OUTLINE_RADIUS);
		glVertex2f(staticObjects[objectIndex].position.x + staticObjects[objectIndex].width, staticObjects[objectIndex].position.y + staticObjects[objectIndex].height - STATIC_OUTLINE_RADIUS);
		glVertex2f(staticObjects[objectIndex].position.x + staticObjects[objectIndex].width, staticObjects[objectIndex].position.y + staticObjects[objectIndex].height);
		glVertex2f(staticObjects[objectIndex].position.x, staticObjects[objectIndex].position.y + staticObjects[objectIndex].height);
		
		glColor4f(1.0f, 0.5f, 0.0f, 1.0f);
		glVertex2f(staticObjects[objectIndex].position.x + STATIC_OUTLINE_RADIUS, staticObjects[objectIndex].position.y + STATIC_OUTLINE_RADIUS);
		glVertex2f(staticObjects[objectIndex].position.x + staticObjects[objectIndex].width - STATIC_OUTLINE_RADIUS, staticObjects[objectIndex].position.y + STATIC_OUTLINE_RADIUS);
		glVertex2f(staticObjects[objectIndex].position.x + staticObjects[objectIndex].width - STATIC_OUTLINE_RADIUS, staticObjects[objectIndex].position.y + staticObjects[objectIndex].height - STATIC_OUTLINE_RADIUS);
		glVertex2f(staticObjects[objectIndex].position.x + STATIC_OUTLINE_RADIUS, staticObjects[objectIndex].position.y + staticObjects[objectIndex].height - STATIC_OUTLINE_RADIUS);
	}
	
	glColor4f(0.5f, 1.0f, 0.5f, 1.0f);
	glVertex2f(movingObject.position.x, movingObject.position.y);
	glVertex2f(movingObject.position.x + movingObject.width, movingObject.position.y);
	glVertex2f(movingObject.position.x + movingObject.width, movingObject.position.y + movingObject.height);
	glVertex2f(movingObject.position.x, movingObject.position.y + movingObject.height);
	
	glEnd();
	
	glutSwapBuffers();
	glutPostRedisplay();
}

static void reshapeFunc(int newWidth, int newHeight) {
	windowWidth = newWidth;
	windowHeight = newHeight;
	glViewport(0, 0, newWidth, newHeight);
}

static void keyDownFunc(int key, int x, int y) {
	switch (key) {
		case GLUT_KEY_LEFT:
			movingObject.moveX = -1;
			break;
		case GLUT_KEY_RIGHT:
			movingObject.moveX = 1;
			break;
		case GLUT_KEY_DOWN:
			movingObject.moveY = -1;
			break;
		case GLUT_KEY_UP:
			movingObject.moveY = 1;
			break;
	}
}

static void keyUpFunc(int key, int x, int y) {
	switch (key) {
		case GLUT_KEY_LEFT:
			if (movingObject.moveX == -1) {
				movingObject.moveX = 0;
			}
			break;
		case GLUT_KEY_RIGHT:
			if (movingObject.moveX == 1) {
				movingObject.moveX = 0;
			}
			break;
		case GLUT_KEY_DOWN:
			if (movingObject.moveY == -1) {
				movingObject.moveY = 0;
			}
			break;
		case GLUT_KEY_UP:
			if (movingObject.moveY == 1) {
				movingObject.moveY = 0;
			}
			break;
	}
}

int main(int argc, char ** argv) {
#ifdef __APPLE__
	GLint VBL = 1;
#endif
	
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE);
	glutInitWindowPosition(0, 0);
	glutInitWindowSize(windowWidth, windowHeight);
	glutCreateWindow("Collision Sample");
	glutReshapeFunc(reshapeFunc);
	glutDisplayFunc(displayFunc);
	glutSpecialFunc(keyDownFunc);
	glutSpecialUpFunc(keyUpFunc);
#ifdef __APPLE__
	CGLSetParameter(CGLGetCurrentContext(), kCGLCPSwapInterval, &VBL);
#endif
	
	FixedIntervalRunLoop_init(&runLoop, getCurrentTime, 1.0f / 120.0f, run);
	initObjects();
	
	glutMainLoop();
	
	return EXIT_SUCCESS;
}
